import importlib
import time

import db
from config import PLUGINS, ALIASES, RESPONSES, BOTS_LOOKUP, SLEEP_TIME
from command_system import RollbotResponse, RollbotFailure
from util import post_message


class Rollbot:
    def __init__(self, logger):
        self.logger = logger
        self.commands = {}
        self.logger.info("Loading command plugins")
        for module, classes in PLUGINS.items():
            plugin_module = importlib.import_module("plugins." + module)
            for class_name in classes:
                plugin_class = getattr(plugin_module, class_name)
                plugin_instance = plugin_class(logger)
                if plugin_instance.command in self.commands:
                    self.logger.error(f"Duplicate command word '{plugin_instance.command}'")
                    raise ValueError(f"Duplicate command word '{plugin_instance.command}'")
                self.commands[plugin_instance.command] = plugin_instance
        self.logger.info(f"Finished loading plugins, {len(self.commands)} commands found")
        self.logger.info("Initializing database tables")
        db.init_db()
        self.logger.info("Finished initializing database")

    def start_plugins(self):
        self.logger.info("Starting all plugins")
        with db.session_scope() as session:
            for plugin_instance in self.commands.values():
                plugin_instance.on_start(session)
        self.logger.info("Finished starting plugins")

    def shutdown_plugins(self):
        self.logger.info("Shutting down all plugins")
        with db.session_scope() as session:
            for plugin_instance in self.commands.values():
                plugin_instance.on_shutdown(session)
        self.logger.info("Finished shutting down plugins")

    def run_command(self, message):
        if not message.is_command:
            self.logger.warn(f"Tried to run non-command message {message.message_id}")
            return RollbotResponse(msg, failure=RollbotFailure.INTERNAL_ERROR)

        # if this command is aliased, resolve that first, otherwise use the literal command
        cmd = ALIASES.get(message.command, message.command)

        if cmd in RESPONSES:
            return RollbotResponse(message, txt=RESPONSES[cmd])

        plugin = self.commands.get(cmd, None)

        if plugin is None:
            self.logger.warn(f"Message {message.message_id} had a command {message.command} (resolved to {cmd}) that could not be run.")
            return RollbotResponse(message, failure=RollbotFailure.INVALID_COMMAND)

        with db.session_scope() as session:
            response = plugin.on_command(session, message)

        if not response.is_success:
            self.logger.warn(f"Message {message.message_id} caused failure")
            self.logger.warn(response.info)

        return response

    def handle_message(self, msg):
        self.logger.info(f"Handling message {msg.message_id}")
        t = time.time()
        try:
            response = self.run_command(msg)
        except Exception as e:
            self.logger.error(f"Exception during command execution {e}, for message {msg.message_id}")
            response = RollbotResponse(msg, failure=RollbotFailure.INTERNAL_ERROR)

        if not response.respond:
            self.logger.info(f"Skipping response to message {msg.message_id}")
            return

        self.logger.info(f"Responding to message {msg.message_id}")

        sleep = SLEEP_TIME - time.time() + t
        if sleep > 0:
            self.logger.info(f"Sleeping for {sleep:.3f}s before responding")
            time.sleep(sleep)

        bot_id = BOTS_LOOKUP[msg.group_id]
        if response.is_success:
            if response.txt is not None:
                post_message(response.txt, bot_id)
            if response.img is not None:
                post_message(response.img, bot_id)
        else:
            post_message(response.failure_msg, bot_id)
            self.logger.warning(f"Failed command response: {response}")

        t = time.time() - t
        self.logger.info(f"Exiting command thread for {msg.message_id} after {t:.3f}s")
